

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2008 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <vector>
#include <map>

#include "plink.h"
#include "options.h"
#include "helper.h"

extern ofstream LOG;
extern Plink * PP;

using namespace std;

void scoreRanges(int, 
		 vector<int> &,
		 map<int, set<int> > &,
		 vector<string> &,
		 vector<int> &,
		 ofstream &);


void Plink::scoreIndividuals()
{
  
  printLOG("Reading set of predictors from [ " + par::score_risk_file + " ]\n");

  checkFileExists(par::score_risk_file);
  ifstream PROFIN;
  PROFIN.open( par::score_risk_file.c_str(), ios::in );

  string problems;
  
  map<string,int> mlocus;
  for(int l=0; l<nl_all;l++)
    mlocus.insert(make_pair(locus[l]->name,l));


  map<int,double> scores;
  map<int,bool> allele1;

  int cnt1 = 0, cnt2 = 0, cnt3 = 0;

  while ( ! PROFIN.eof() )
    {
      
      // Format assumed: SNP allele score
      
      string snp, allele, sscore;
      double score;
      
      PROFIN >> snp >> allele >> sscore;

      if ( sscore=="" )
	continue;

      if ( ! from_string<double>( score, sscore , std::dec))
	error("Problem reading predictors: non-numeric value found for {SNP} {allele} {value}:\n{"
	      +snp+"} {"+allele+"} {"+sscore+"}");
      
      ++cnt1;

      map<string,int>::iterator ilocus = mlocus.find(snp);
      
      // SNP not found
      if ( ilocus == mlocus.end() )
	{
	  problems += "NOSNP\t" + snp + "\n";
	  continue;
	}

      ++cnt2;

      int l = ilocus->second;

      // Allele not found
      if ( allele == locus[l]->allele1 )
	{
	  scores.insert(make_pair(l,score));
	  allele1.insert(make_pair(l,false));
	  ++cnt3;
	}
      else if ( allele == locus[l]->allele2 )
	{
	  scores.insert(make_pair(l,score));
	  allele1.insert(make_pair(l,true));
	  ++cnt3;
	}
      else
	problems += "NOALLELE\t" 
	  + snp + " " + allele 
	  + " vs " + locus[l]->allele1 
	  + " " + locus[l]->allele2 + "\n";

    }

  PROFIN.close();
  
  printLOG("Read " + int2str(cnt1) 
	   + " predictors; " + int2str(cnt2) 
	   + " mapped to SNPs; " + int2str(cnt3) 
	   + " to alleles\n");

  if ( cnt1 != cnt2 || 
       cnt1 != cnt3 )
    { 
      printLOG("Writing problem SNPs in predictor to [ " 
	       + par::output_file_name + ".nopred ]\n");
      ofstream O1;
      O1.open( (par::output_file_name + ".nopred").c_str() , ios::out );
      O1 << problems ;
      O1.close();
      problems = "";
    }

  


  ////////////////////////////////
  // Calculate for each individual
  
  printLOG("Writing profiles to [ " + par::output_file_name + ".profile ]\n");

  vector_t profile;
  vector<int> cnt;
  vector<int> acount;
  
  calculateProfile(scores,allele1,profile,cnt,acount);



  ///////////////////////////////
  // Report for all individuals

  ofstream PROFOUT;
  string f = par::output_file_name + ".profile";
  PROFOUT.open( f.c_str(), ios::out );

  PROFOUT << setw(par::pp_maxfid) << "FID" << " " 
	  << setw(par::pp_maxiid) << "IID" << " "
	  << setw(6) << "PHENO" << " " 
	  << setw(6) << "CNT" << " "
          << setw(6) << "CNT2" << " "
	  << setw(8) << "SCORE" << "\n";
  
  for ( int i=0; i<n; i++ )
    {
      Individual * person = sample[i];

      PROFOUT << setw(par::pp_maxfid) << person->fid << " " 
	      << setw(par::pp_maxiid) << person->iid << " "
	      << setw(6) << person->phenotype << " " 
	      << setw(6) << cnt[i] << " "
	      << setw(6) << acount[i] << " " 
	      << setw(8) << profile[i] << "\n";
    }
  
  PROFOUT.close();

}


void Plink::calculateProfile(map<int,double> & scores, 
			     map<int,bool> & allele1, 
			     vector_t & profile,
			     vector<int> & count,
			     vector<int> & acount )
{  
  

  // Generate a vector of scores, one for each individual, given then
  // scoring set (and allele direction) for a set of SNPs

  profile.resize(n,0);
  count.resize(n,0);
  acount.resize(n,0);


  ///////////////////////////////////////
  // Do we want to score for genes also?

  map<string, set<Range> > ranges;
  vector<string> rangeLabels;
  map<int,set<int> > snp2range;
  vector<int> rangeCount;
  ofstream ROUT;

  if ( par::score_risk_ranges )
    {
      printLOG("Writing range scores to [ " 
	       + par::output_file_name 
	       + ".profile.ranges ]\n");

      // Helper function to map ranges to SNPs
      mapRangesToSNPs( par::score_risk_ranges_file,
		       ranges, 
		       rangeLabels,
		       snp2range );

      rangeCount.resize(rangeLabels.size());

      ROUT.open( ( par::output_file_name+".profile.ranges").c_str(), ios::out);
   
      // Header
      
      ROUT << setw(par::pp_maxfid) << "FID" << " " 
	   << setw(par::pp_maxiid) << "IID" << " "
	   << setw(6) << "PHENO" << " ";

      for (int r=0; r<rangeLabels.size(); r++)
	ROUT << rangeLabels[r] << " ";
      ROUT << "\n";
      	      	   
    }
  

  //////////////////////////////////////////
  // Consider each individual and calculate
  // the score
  
  for (int i=0; i<n; i++)
    {

      Individual * person = sample[i];
      
      map<int,double>::iterator i1 = scores.begin();
      map<int,bool>::iterator i2 = allele1.begin();
      
      double score = 0;
      int cnt = 0;
      int cntActual = 0;
      int cntNamedAllele = 0;
	  
      vector<int> flaggedSNPs;
      
      while ( i1 != scores.end() )
	{
	  
	  int l = i1->first;
	  bool a1 = i2->second;

	  bool s1 = par::SNP_major ? SNP[l]->one[i] : person->one[l];
	  bool s2 = par::SNP_major ? SNP[l]->two[i] : person->two[l];

	  bool missingGenotype = false;

	  double thisScore = 0;

	  /////////////////////////////////////////////
	  // Individual is missing this genotype
	  
	  // We with either skip, or impute mean

	  if ( s1 && ! s2 ) 
	    {
	      
	      if ( ! par::score_impute_expected ) 
		{
		  ++i1;
		  ++i2;
		  continue;
		}
	      
	      missingGenotype = true;
	      
	      if ( i2->second ) 
		thisScore = ( 1 - locus[l]->freq ) * i1->second;
	      else
		thisScore = locus[l]->freq * i1->second;
	      
	      if ( par::chr_haploid[ locus[l]->chr ] || 
		   ( par::chr_sex[ locus[l]->chr ] && person->sex ) )
		++cnt;
	      else
		{
		  cnt += 2;
		  thisScore *= 2;
		}
	    }


	  // Currently, just an allelic scoring: we could extend this 
	  // to genotypes, dominant/recessive models,
	  
	  bool sawNamedAllele = false;
	  
	  if ( ! missingGenotype ) 
	    { 
	      
	      if ( par::chr_haploid[ locus[l]->chr ] || 
		   ( par::chr_sex[ locus[l]->chr ] && person->sex ) )
		{
		  // A single copy
		  
		  if ( i2->second ) 
		    {
		      if ( s1 ) 
			{
			  thisScore = i1->second;
			  sawNamedAllele = true;
			  ++cntNamedAllele;
			}
		    }
		  else
		    {
		      if ( !s1 ) 
			{
			  thisScore = i1->second;
			  sawNamedAllele = true;
			  ++cntNamedAllele;
			}
		    }
		  
		  ++cnt;	      
		  ++cntActual;
		}
	      else // .. autosomal
		{
		  if ( i2->second ) 
		    {
		      if ( s1 ) 
			{
			  thisScore = i1->second;
			  sawNamedAllele = true;
			  ++cntNamedAllele;
			}
		      if ( s2 ) 
			{
			  thisScore += i1->second;
			  sawNamedAllele = true;
			  ++cntNamedAllele;
			}
		    }
		  else
		    {
		      if ( !s1 ) 
			{
			  thisScore = i1->second;
			  sawNamedAllele = true;
			  ++cntNamedAllele;
			}
		      if ( !s2 ) 
			{
			  thisScore += i1->second;	      
			  sawNamedAllele = true;
			  ++cntNamedAllele;
			}
		    }
		  cnt += 2;	      
		  cntActual +=2;
		}
	    }
	  
	  //////////////////////////////////////////
	  // Accumulate score
	  
	  score += thisScore;
	  

	  //////////////////////////////////////////
	  // Do we want to score a "yes/no" for a
	  // gene?

	  if ( par::score_risk_ranges && ! missingGenotype )
	    {

	      // Did we see at least one risk-increasing allele?
	      
	      if ( ( sawNamedAllele && i1->second > 0 ) ||
		   ( (!sawNamedAllele) && i1->second < 0 ) ) 
		{
		  flaggedSNPs.push_back(l);
		}
	    }
	  
	  ++i1;
	  ++i2;

	}
      
      
      // Get average per seen loci
      
      if ( cnt>0 ) 
	score /= (double)cnt;

      // Save for this individual (actual number)

      profile[i] = score;
      count[i] = cntActual;
      acount[i] = cntNamedAllele;

      // Score for ranges?
      
      if ( par::score_risk_ranges )
	{
	  scoreRanges(i,flaggedSNPs,snp2range,rangeLabels,rangeCount,ROUT);
	}

     
    } // Next individual
  

  if ( par::score_risk_ranges )
    {
      ROUT.close();
      
      // Also ouput how many times each range was seen
      
      printLOG("Writing range summary counts to [ "
	       + par::output_file_name 
	       + ".profile.ranges.summary ]\n");
      
      ofstream ROUT2;
      ROUT2.open( ( par::output_file_name+".profile.ranges.summary").c_str(), ios::out);
   
      // Header
      
      ROUT2 << setw(18) << "RANGE" << " " 
	   << setw(8) << "CNT" << "\n";
      for (int r=0; r<rangeLabels.size(); r++)
	ROUT2 << setw(18) << rangeLabels[r] << " "
	      << setw(8) << rangeCount[r] << "\n";
      ROUT2.close();
    }
    
  return;
}


void scoreRanges(int i, 
		 vector<int> & f,		 
		 map<int, set<int> > & snp2range,
		 vector<string> & labels,
		 vector<int> & rangeCount,
		 ofstream & ROUT)
{
  Individual * person = PP->sample[i];
  
  ROUT << setw(par::pp_maxfid) << person->fid << " "
       << setw(par::pp_maxiid) << person->iid << " "
       << setw(6) << person->phenotype << " ";
  
  // Get list of ranges that will be flagged

  set<int> mappedRanges;

  for (int l=0; l<f.size(); l++)
    {
      map<int, set<int> >::iterator ri = snp2range.find( f[l] );
      
      if ( ri == snp2range.end() )
	continue;

      set<int>::iterator si = ri->second.begin();
      while ( si != ri->second.end() )
	{
	  mappedRanges.insert( *si );
	  ++si;
	}
    }
  
  // We now have populated the set mappedRanges

  for (int r=0; r<labels.size(); r++)
    {
      if ( mappedRanges.find(r) != mappedRanges.end() )
	{
	  ROUT << "1 ";
	  ++rangeCount[r];
	}
      else
	ROUT << "0 ";
    }
  ROUT << "\n";

}
