

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2008 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>

#include "plink.h"
#include "helper.h"
#include "options.h"
#include "crandom.h"

using namespace std;


//////////////////////////////////////////////////////////////////////
// A simple routine to simulate a dataset of unlinked case/control SNPs

class SimParameters
{
public:
  int nsnp;
  double lfreq;
  double ufreq;
  double odds;
  double missing;
  string name;

  double lmarker;
  double umarker;
  double dprime;

  SimParameters()
  {
    name = "";
    nsnp = 0;
    missing = 0.00;
    lfreq = ufreq = odds = 0;
    lmarker = umarker = 0;
    dprime = 1;
  }
};


vector_t instanceSNP(SimParameters & s)
{
  
  // Return:
  // 0 Population allele frequency (disease variant)
  // 1 Population allele frequency (marker)
  // 2 Case AA (marker)
  // 3 Case AB (marker)
  // 4 Control AA (marker)
  // 5 Control AB (marker)

  
  vector_t freqs(6,0);
  
  // Calculate actual population allele frequency for this SNP
  
  double freq = s.lfreq + CRandom::rand() * ( s.ufreq - s.lfreq ) ;
  
  // And a marker frequency

  double mfreq = par::simul_tags ? 
    s.lmarker + CRandom::rand() * ( s.umarker - s.lmarker ) :
    freq;


  // Handle LD

  double dmax;
  double ld = 0;
  double h11, h12, h21, h22;

  dmax = freq * (1-mfreq);
  if ( (1-freq) * mfreq < dmax) dmax = (1-freq) * mfreq;
  ld = s.dprime * dmax;
   
  h11 = freq * mfreq + ld;
  h12 = freq * ( 1 - mfreq ) - ld;
  h21 = ( 1 - freq ) * mfreq - ld;
  h22 = ( 1 - freq ) * ( 1 - mfreq ) + ld;

  // Get case and control allele frequencies given GRR, 
  // population frequency and disease frequency

  double f0, f1, f2;
  double g0 = freq * freq;
  double g1 = 2 * freq * ( 1-freq );
  double g2 = 1 - g0 - g1;

  f2 = par::simul_prevalence / ( g0 * s.odds * s.odds + g1 * s.odds + g2 );
  f0 = f2 * s.odds * s.odds;
  f1 = f2 * s.odds;

  // Calculate risk at marker locus

  double mg0 = mfreq * mfreq;
  double mg1 = 2 * mfreq * ( 1 - mfreq);
  double mg2 = 1 - mg0 - mg1;

  double mf0 = ( f0 * h11*h11
		+ f1 * 2 * h11 * h21 
		+ f2 * h21*h21 ) / mg0;
  
  double mf1 = ( f0 * 2*h11*h12
		+ f1 * ( 2 * h11 * h22 + 2 * h12 * h21 ) 
		+ f2 * 2*h21*h22 ) / mg1;
  
  double mf2 = ( f0 * h12*h12
		+ f1 * 2 * h12 * h22 
		+ f2 * h22*h22 ) / mg2;
  
  
  // P(G|X)
  
  double d0 = mg0 * mf0;
  double d1 = mg1 * mf1;
  double d2 = mg2 * mf2;
  double dS = d0 + d1 + d2;
  d0 /= dS;
  d1 /= dS;
  d2 /= dS;

  double u0 = mg0 * (1-mf0);
  double u1 = mg1 * (1-mf1);
  double u2 = mg2 * (1-mf2);
  double uS = u0 + u1 + u2;
  u0 /= uS;
  u1 /= uS;
  u2 /= uS;

  freqs[0] = freq;
  freqs[1] = mfreq;

  freqs[2] = d0;
  freqs[3] = d1;

  freqs[4] = u0;
  freqs[5] = u1;

  return freqs;
}


void Plink::simulateSNPs()
{

  // Read in SNP parameters

  // Number of SNPs
  // Lower allele frequency for '1' (versus '2') allele
  // Upper allele frequency (population)
  // Odds ratio ('1' allele)

  checkFileExists(par::simul_file);
  printLOG("Reading simulation parameters from [ " 
	   + par::simul_file + " ]\n");
  
  printLOG("Writing SNP population frequencies to [ " 
	   + par::output_file_name + ".simfreq ]\n");
  
  ofstream SOUT( ( par::output_file_name+".simfreq").c_str(), ios::out);
  
  ifstream SIM;
  SIM.open( par::simul_file.c_str(), ios::in );
  
  vector<SimParameters> sp;

  while ( ! SIM.eof() )
    {
      SimParameters s;

      if ( par::simul_tags )
	{
	  SIM >> s.nsnp
	      >> s.name
	      >> s.lfreq 
	      >> s.ufreq 
	      >> s.lmarker
	      >> s.umarker
	      >> s.dprime
	      >> s.odds;
	}
      else
	{      
	  SIM >> s.nsnp
	      >> s.name
	      >> s.lfreq 
	      >> s.ufreq 
	      >> s.odds;
	  s.lmarker = s.lfreq;
	  s.umarker = s.ufreq;
	  s.dprime = 1;
	}

      if ( SIM.eof() )
	continue;

      sp.push_back(s);

    }

  SIM.close();


  ////////////////////////////////////////////
  // Make room for total number of SNPs, etc
  
  int tsnp = 0;
  
  for (int s=0; s<sp.size(); s++)
    tsnp += sp[s].nsnp;
  
  int nind = par::simul_ncases + par::simul_ncontrols;
      
  printLOG("Read " + int2str( sp.size() ) 
	   + " sets of SNPs, specifying " 
	   + int2str(tsnp) 
	   + " SNPs in total\n");
  printLOG("Simulating " 
	   + int2str( par::simul_ncases ) 
	   + " cases and " 
	   + int2str( par::simul_ncontrols ) 
	   + " controls\n");
  printLOG("Assuming a disease prevalence of " 
	   + dbl2str( par::simul_prevalence ) + "\n");
  
  if ( par::simul_tags )
    printLOG("Simulating markers in indirect association\n\n");
  else
    printLOG("Simulating disease variants (direct association)\n\n");


  int pos = 0;

  for (int s=0; s<sp.size(); s++)  
    for (int l=0;l<sp[s].nsnp;l++)
    {
      Locus * loc = new Locus;
      
      // Optionally add
      if ( sp[s].nsnp > 1 ) 
	loc->name = sp[s].name+"_"+int2str(l);
      else
	loc->name = sp[s].name;

      loc->chr = 1;
      loc->allele1 = "A";
      loc->allele2 = "B";
      loc->bp = ++pos; 
      loc->pos = 0;

      locus.push_back(loc);

      CSNP * newset = new CSNP;
      
      newset->one.resize(nind);
      newset->two.resize(nind);

      // Sample case and control population allele frequencies

      // Contains population, case, control frequencies
      
      vector_t f = instanceSNP(sp[s]);      
      
      // f Information

      // 0 Population allele frequency (disease variant/marker)
      // 1 Population allele frequency (marker)
      // 2 Case AA (marker)
      // 3 Case AB (marker)
      // 4 Control AA (marker)
      // 5 Control AB (marker)
      
      if ( par::simul_tags )
	{
	  SOUT << 1 << " "
	       << loc->name << "\t"
	       << f[0] << " " << f[0] << "\t" 
	       << f[1] << " " << f[1] << "\t" 
	       << sp[s].dprime << "\t"
	       << sp[s].odds << "\n";
	}
      else
	{	  
	  SOUT << 1 << " "
	       << loc->name << "\t"
	       << f[0] << " " << f[0] << "\t" 
	       << sp[s].odds << "\n";	  
	}


      // Genotype frequencies in cases and controls

      const double caseAA = f[2];
      const double caseAB = f[2] + f[3];
      
      const double contAA = f[4];
      const double contAB = f[4] + f[5];
      

      //////////////////////////////////////////////////
      // Generate each individual, simulating genotypes
      // rather than alleles
      
      for ( int i = 0 ; i < nind ; i++ ) 
	{
	  
	  // Simple missingness
	  
	  if ( CRandom::rand() < sp[s].missing ) 
	    {
	      newset->one[i] = true;
	      newset->two[i] = false;
	    }
	  else
	    {
	      
	      bool isCase = i < par::simul_ncases ? true : false;
	      
	      double r = CRandom::rand();
	      
	      int g = 0;
	      
	      if ( isCase ) 
		{
		  if ( r > caseAB )
		    g = 2;		  
		  else if ( r > caseAA )
		    g = 1;
		}
	      else
		{
		  if ( r > contAB )
		    g = 2;
		  else if ( r > contAA )
		    g = 1;
		  
		}
	      
	      
	      if ( g == 2 ) 
		{
		  newset->one[i] = false;
		  newset->two[i] = false;
		}
	      else if ( g == 1 ) 
		{
		  newset->one[i] = false;
		  newset->two[i] = true;
		}
	      else
		{
		  newset->one[i] = true;
		  newset->two[i] = true;
		}
	      
	    }
	}
      
      SNP.push_back(newset);
    }
  
  
  // Phenotypes

  for (int i=0;i<nind;i++)
    {
      Individual * person = new Individual;
      person->fid = person->iid = "per"+int2str(i);
      person->missing = false;
      person->pat = "0";
      person->mat = "0";

      if ( i < par::simul_ncases ) 
	person->phenotype = 2;
      else
	person->phenotype = 1;

      person->sex = false;
      person->sexcode = "2";
      sample.push_back(person);      
    }
  
  SOUT.close();


}

